@@ -1,5 +1,5 @@
 import sys
-from typing import List, Dict, Any, Callable, Optional
+from typing import List, Dict, Any, Callable, Optional, Tuple, Sequence, Union
 from typing import Type as PyType
 from typing import cast
 
@@ -15,6 +15,7 @@ from mypy.types import (
     FunctionLike,
 )
 from mypy.checker import TypeChecker, is_false_literal
+from mypy.options import Options
 from mypy.nodes import TypeInfo
 from mypy.plugin import (
     CheckerPluginInterface,
@@ -28,12 +29,15 @@ from mypy.plugin import (
     ClassDefContext,
 )
 from mypy.subtypes import find_member
+from mypy.plugins.default import DefaultPlugin
 
 from mypy.nodes import (
     Context,
     Var,
     Argument,
     FuncDef,
+    OverloadedFuncDef,
+    Decorator,
     CallExpr,
     RefExpr,
     Expression,
@@ -49,6 +53,8 @@ from mypy.nodes import (
     ARG_OPT,
     FUNC_NO_INFO,
     NameExpr,
+    MypyFile,
+    SymbolNode,
 )
 
 from collections import defaultdict
@@ -91,6 +97,10 @@ SIMPLE_FIELD_TO_TYPE = {
 
 
 class ZopeInterfacePlugin(Plugin):
+    def __init__(self, options: Options):
+        super().__init__(options)
+        self.fallback = DefaultPlugin(options)
+
     def log(self, msg: str) -> None:
         if self.options.verbosity >= 1:
             print("ZOPE:", msg, file=sys.stderr)
@@ -105,6 +115,7 @@ class ZopeInterfacePlugin(Plugin):
         self, fullname: str
     ) -> Optional[Callable[[FunctionContext], Type]]:
         # print(f"get_function_hook: {fullname}")
+
         def analyze(function_ctx: FunctionContext) -> Type:
             # strtype = function_ctx.api.named_generic_type('builtins.str', [])
             # optstr = function_ctx.api.named_generic_type('typing.Optional', [strtype])
@@ -121,6 +132,60 @@ class ZopeInterfacePlugin(Plugin):
 
             return deftype
 
+        def class_implements_hook(function_ctx: FunctionContext) -> Type:
+            assert len(function_ctx.arg_types) == 2
+            arg_impl = function_ctx.arg_types[0][0]
+            expr = function_ctx.args[0][0]
+            implname = expr.fullname if isinstance(expr, NameExpr) else "expression"
+            if not isinstance(arg_impl, CallableType):
+                function_ctx.api.fail(
+                    f"{implname} is not a class, "
+                    "cannot mark it as a interface implementation",
+                    function_ctx.context,
+                )
+                return function_ctx.default_return_type
+            assert isinstance(arg_impl, CallableType)
+            impl_type = arg_impl.ret_type
+            assert isinstance(impl_type, Instance)
+
+            for expr, arg_iface in zip(function_ctx.args[1], function_ctx.arg_types[1]):
+                exprname = expr.fullname if isinstance(expr, NameExpr) else "expression"
+                if not isinstance(arg_iface, CallableType):
+                    function_ctx.api.fail(
+                        f"{exprname} is not a class, "
+                        f"cannot mark {implname} as an implementation of "
+                        f"{exprname}",
+                        function_ctx.context,
+                    )
+                    continue
+
+                iface_type = arg_iface.ret_type
+                assert isinstance(iface_type, Instance)
+                if not self._is_interface(iface_type.type):
+                    function_ctx.api.fail(
+                        f"{exprname} is not an interface", function_ctx.context,
+                    )
+                    function_ctx.api.fail(
+                        f"Make sure you have stubs for all packages that "
+                        f"provide interfaces for {exprname} class hierarchy.",
+                        function_ctx.context,
+                    )
+                    continue
+
+                self._apply_interface(impl_type.type, iface_type.type)
+                self._report_implementation_problems(
+                    impl_type, iface_type, function_ctx.api, function_ctx.context
+                )
+
+            return function_ctx.default_return_type
+
+        if fullname == "zope.interface.declarations.classImplements":
+            return class_implements_hook
+
+        # Give preference to deault plugin
+        hook = self.fallback.get_function_hook(fullname)
+        if hook is not None:
+            return hook
         return analyze
 
     def get_method_signature_hook(
@@ -190,11 +255,10 @@ class ZopeInterfacePlugin(Plugin):
     ) -> Optional[Callable[[ClassDefContext], None]]:
         # print(f"get_class_decorator_hook: {fullname}")
 
-        def apply_interface(
+        def apply_implementer(
             iface_arg: Expression,
             class_info: TypeInfo,
             api: SemanticAnalyzerPluginInterface,
-            context: Context,
         ) -> None:
             if not isinstance(iface_arg, RefExpr):
                 api.fail(
@@ -226,22 +290,7 @@ class ZopeInterfacePlugin(Plugin):
                 )
                 return
 
-            # print("CLASS INFO", class_info)
-            md = self._get_metadata(class_info)
-            if "implements" not in md:
-                md["implements"] = []
-            # impl_list = cast(List[str], md['implements'])
-            md["implements"].append(iface_type.fullname)
-            self.log(
-                f"Found implementation of "
-                f"{iface_type.fullname}: {class_info.fullname}"
-            )
-
-            # Make sure implementation is treates subtype of an interface. Pretend
-            # there is a decorator for the class that will create a "type promotion"
-            faketi = TypeInfo(SymbolTable(), iface_type.defn, iface_type.module_name)
-            faketi._promote = Instance(iface_type, [])
-            class_info.mro.append(faketi)
+            self._apply_interface(class_info, iface_type)
 
         def analyze(classdef_ctx: ClassDefContext) -> None:
             api = classdef_ctx.api
@@ -249,7 +298,7 @@ class ZopeInterfacePlugin(Plugin):
             decor = cast(CallExpr, classdef_ctx.reason)
 
             for iface_arg in decor.args:
-                apply_interface(iface_arg, classdef_ctx.cls.info, api, decor)
+                apply_implementer(iface_arg, classdef_ctx.cls.info, api)
 
         if fullname == "zope.interface.declarations.implementer":
             return analyze
@@ -426,10 +475,13 @@ class ZopeInterfacePlugin(Plugin):
             return
 
         for idx, item in enumerate(cls.defs.body):
-            if not isinstance(item, FuncDef):
+            if isinstance(item, FuncDef):
+                replacement = self._adjust_interface_function(api, cls.info, item)
+            elif isinstance(item, OverloadedFuncDef):
+                replacement = self._adjust_interface_overload(api, cls.info, item)
+            else:
                 continue
 
-            replacement = self._adjust_interface_function(api, cls.info, item)
             cls.defs.body[idx] = replacement
 
         md["interface_analyzed"] = True
@@ -470,6 +522,23 @@ class ZopeInterfacePlugin(Plugin):
 
         return func_def
 
+    def _adjust_interface_overload(
+        self,
+        api: SemanticAnalyzerPluginInterface,
+        class_info: TypeInfo,
+        overload_def: OverloadedFuncDef,
+    ) -> Statement:
+
+        for overload_item in overload_def.items:
+            if isinstance(overload_item, Decorator):
+                func_def = overload_item.func
+            else:
+                assert isinstance(overload_item, FuncDef)
+                func_def = overload_item
+
+            self._adjust_interface_function(api, class_info, func_def)
+        return overload_def
+
     def _report_implementation_problems(
         self,
         impl_type: Instance,
@@ -503,7 +572,7 @@ class ZopeInterfacePlugin(Plugin):
 
         if missing:
             for iface_name, members in missing.items():
-                missing_fmt = ", ".join(members)
+                missing_fmt = ", ".join(sorted(members))
                 api.fail(
                     f"'{impl_info.name}' is missing following "
                     f"'{iface_name}' interface members: {missing_fmt}.",
@@ -575,18 +644,31 @@ class ZopeInterfacePlugin(Plugin):
                 pass
 
     def _lookup_type(self, fullname: str, api: TypeChecker) -> Instance:
-        # Implement own type lookup, because TypeChecker.named_generic_type is
-        # dysfunctional
-        parts = fullname.split(".")
-        module_name = ".".join(parts[:-1])
-        type_name = parts[-1]
-        module = api.modules[module_name]
-        sym = module.names[type_name]
-        typeinfo = sym.node
+        module, names = self._find_module(fullname, api)
+        container: Union[MypyFile, TypeInfo] = module
+        for name in names:
+            sym = container.names[name]
+            assert isinstance(sym.node, TypeInfo)
+            container = sym.node
+        typeinfo = container
         assert isinstance(typeinfo, TypeInfo)
+
         any_type = AnyType(TypeOfAny.from_omitted_generics)
         return Instance(typeinfo, [any_type] * len(typeinfo.defn.type_vars))
 
+    def _find_module(
+        self, fullname: str, api: TypeChecker
+    ) -> Tuple[MypyFile, Sequence[str]]:
+        parts = fullname.split(".")
+        for pn in range(len(parts) - 1, 0, -1):
+            moduleparts = parts[:pn]
+            names = parts[pn:]
+            modulename = ".".join(moduleparts)
+            if modulename in api.modules:
+                return api.modules[modulename], names
+
+        raise LookupError(fullname)
+
     def _index_members(self, typeinfo: TypeInfo) -> Dict[str, TypeInfo]:
         # we skip "object" and "Interface" since everyone implements it
         members = {}
@@ -600,6 +682,24 @@ class ZopeInterfacePlugin(Plugin):
                 members[name] = base
         return members
 
+    def _apply_interface(self, impl: TypeInfo, iface: TypeInfo) -> None:
+
+        md = self._get_metadata(impl)
+        if "implements" not in md:
+            md["implements"] = []
+
+        md["implements"].append(iface.fullname)
+        self.log(f"Found implementation of " f"{iface.fullname}: {impl.fullname}")
+
+        # Make sure implementation is treated as a subtype of an interface. Pretend
+        # there is a decorator for the class that will create a "type promotion",
+        # but ensure this only gets applied a single time per interface.
+        promote = Instance(iface, [])
+        if not any(ti._promote == promote for ti in impl.mro):
+            faketi = TypeInfo(SymbolTable(), iface.defn, iface.module_name)
+            faketi._promote = promote
+            impl.mro.append(faketi)
+
 
 def plugin(version: str) -> PyType[Plugin]:
     return ZopeInterfacePlugin
